// license:BSD-3-Clause
// copyright-holders:FIND
////////////////////////////////////////////////////////////////////////////////////////////////////
//
// class DisasmWindow
// tAZuXg\NX
//
////////////////////////////////////////////////////////////////////////////////////////////////////
#include "stdafx.h"
#include <vector>
#include "VJR200.h"
#include "JRSystem.h"
#include "DisasmWindow.h"

static const LPCWSTR szWindowClassMem = L"Disassemble Window";
static const int MAXLINES = 4096;
static int lineHeight = 16;
static int dispno, pos;

extern JRSystem sys;

enum class AddressingMode { IMMEDIATE = 0, DIRECT = 1, INDEX = 2, EXTEND = 3, IMPLIED = 4, RELATIVEAD = 5 };

DisasmWindow::DisasmWindow(): hwnd(NULL),
pRenderTarget(NULL),
pTextFormat(NULL),
pBlackBrush(NULL),
pTextLayout(NULL)
{
}

DisasmWindow::~DisasmWindow()
{
	SafeRelease(&pTextFormat);
	SafeRelease(&pTextLayout);

	DiscardDeviceResources();
}

////////////////////////////////////////////////////////////////////////////////////////////////////
//
// 
//
////////////////////////////////////////////////////////////////////////////////////////////////////
HRESULT DisasmWindow::Init()
{
	HRESULT hr;
	hr = CreateDeviceIndependentResources();

	if (SUCCEEDED(hr)) {
		WNDCLASSEX wcex;

		wcex.cbSize = sizeof(WNDCLASSEX);

		wcex.style = CS_HREDRAW | CS_VREDRAW | CS_DBLCLKS;
		wcex.lpfnWndProc = WndProcDisasmWindow;
		wcex.cbClsExtra = 0;
		wcex.cbWndExtra = 0;
		wcex.hInstance = g_hInst;
		wcex.hIcon = NULL;
		wcex.hCursor = LoadCursor(NULL, IDC_ARROW);
		wcex.hbrBackground = (HBRUSH)(COLOR_WINDOW + 1);
		wcex.lpszMenuName = NULL;
		wcex.lpszClassName = szWindowClassMem;
		wcex.hIconSm = LoadIcon(wcex.hInstance, MAKEINTRESOURCE(IDI_SMALL));

		RegisterClassEx(&wcex);

		RECT rect;
		SystemParametersInfo(SPI_GETWORKAREA, 0, &rect, 0);
		hwnd = CreateWindow(
			szWindowClassMem,
			g_stringTable[38],
			WS_OVERLAPPED | WS_SYSMENU | WS_THICKFRAME,
			rect.right - 300,
			0,
			300,
			rect.bottom,
			NULL,
			NULL,
			g_hInst,
			nullptr
		);
		hr = hwnd ? S_OK : E_FAIL;
		if (SUCCEEDED(hr))
		{
			ShowWindow(hwnd, SW_SHOWNA);
		}
	}
	return hr;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
//
// EBhEnhԂ
//
////////////////////////////////////////////////////////////////////////////////////////////////////
HWND DisasmWindow::GetHWnd()
{
	return hwnd;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
//
// EBhE\
//
////////////////////////////////////////////////////////////////////////////////////////////////////
void DisasmWindow::Show()
{
	ShowWindow(hwnd, SW_SHOWNA);

}

////////////////////////////////////////////////////////////////////////////////////////////////////
//
// foCXƗ\[X
//
////////////////////////////////////////////////////////////////////////////////////////////////////
HRESULT DisasmWindow::CreateDeviceIndependentResources()
{
	static const wchar_t msc_fontName[] = L"Courier New";
	static const FLOAT msc_fontSize = 14;
	HRESULT hr;

	// Create a DirectWrite text format object.
	hr = pDWriteFactory->CreateTextFormat(
		msc_fontName,
		NULL,
		DWRITE_FONT_WEIGHT_NORMAL,
		DWRITE_FONT_STYLE_NORMAL,
		DWRITE_FONT_STRETCH_NORMAL,
		msc_fontSize,
		L"", //locale
		&pTextFormat
	);

	if (SUCCEEDED(hr))
	{
		// Center the text horizontally and vertically.
		pTextFormat->SetTextAlignment(DWRITE_TEXT_ALIGNMENT_LEADING);
		pTextFormat->SetParagraphAlignment(DWRITE_PARAGRAPH_ALIGNMENT_NEAR);
	}

	return hr;
}

////////////////////////////////////////////////////////////////////////////////////////////////////
//
// foCX\[X
//
////////////////////////////////////////////////////////////////////////////////////////////////////
HRESULT DisasmWindow::CreateDeviceResources()
{
	HRESULT hr = S_OK;

	if (!pRenderTarget)
	{

		RECT rc;
		GetClientRect(hwnd, &rc);

		D2D1_SIZE_U size = D2D1::SizeU(
			rc.right - rc.left,
			rc.bottom - rc.top
		);

		// Create a Direct2D render target.
		hr = pD2dFactory->CreateHwndRenderTarget(
			D2D1::RenderTargetProperties(),
			D2D1::HwndRenderTargetProperties(hwnd, size),
			&pRenderTarget
		);
		if (SUCCEEDED(hr))
		{
			// Create a black brush.
			hr = pRenderTarget->CreateSolidColorBrush(
				D2D1::ColorF(D2D1::ColorF::Black),
				&pBlackBrush
			);
		}
	}

	return hr;
}


////////////////////////////////////////////////////////////////////////////////////////////////////
//
// DirecWrite`
//
////////////////////////////////////////////////////////////////////////////////////////////////////
HRESULT DisasmWindow::OnRender()
{
	static int prevAddress;
	static SCROLLINFO si;
	HRESULT hr;

	hr = CreateDeviceResources();

	if (SUCCEEDED(hr) && !(pRenderTarget->CheckWindowState() & D2D1_WINDOW_STATE_OCCLUDED))
	{
		// Retrieve the size of the render target.
		D2D1_SIZE_F renderTargetSize = pRenderTarget->GetSize();

		pRenderTarget->BeginDraw();
		pRenderTarget->SetTransform(D2D1::Matrix3x2F::Identity());
		pRenderTarget->Clear(D2D1::ColorF(D2D1::ColorF::White));

		wchar_t  text[8000];
		D2D1_POINT_2F points;
		points.x = 10.0f;
		points.y = 0.0f;
		text[0] = '\0';

		uint16_t pc, sp, x;
		uint8_t a, b, cc;
		sys.pCpu->GetRegister(pc, sp, x, a, b, cc);
		curAddress = pc;
		for (int i = 0; i <  dispno; ++i) {
			OperationSet oSet;
			wchar_t line[40];
			prevAddress = curAddress;
			SetOperation(&oSet);
			if (curAddress - prevAddress < 0) break;
			wsprintf(line, L"%04X %s%s %s\r\n", oSet.startAddress, oSet.bytesString, oSet.nimonic, oSet.operand);
			wcscat(text, line);
		}

		hr = pDWriteFactory->CreateTextLayout(
			text,
			wcslen(text),
			pTextFormat,
			renderTargetSize.width,
			renderTargetSize.height,
			&pTextLayout
		);

		if (SUCCEEDED(hr))
			pRenderTarget->DrawTextLayout(points, pTextLayout, pBlackBrush, D2D1_DRAW_TEXT_OPTIONS_CLIP);
		SafeRelease(&pTextLayout);

		hr = pRenderTarget->EndDraw();

		if (hr == D2DERR_RECREATE_TARGET)
		{
			hr = S_OK;
			DiscardDeviceResources();
		}
	}

	return hr;
}


////////////////////////////////////////////////////////////////////////////////////////////////////
//
// TCY
//
////////////////////////////////////////////////////////////////////////////////////////////////////
void DisasmWindow::OnResize(UINT width, UINT height)
{
	if (pRenderTarget)
	{
		D2D1_SIZE_U size;
		size.width = width;
		size.height = height;

		pRenderTarget->Resize(size);
	}

}


////////////////////////////////////////////////////////////////////////////////////////////////////
//
// EBhE
//
////////////////////////////////////////////////////////////////////////////////////////////////////
void DisasmWindow::Close()
{
	SendMessage(hwnd, WM_CLOSE, 0, 0);
}


////////////////////////////////////////////////////////////////////////////////////////////////////
//
// foCX\[Xp
//
////////////////////////////////////////////////////////////////////////////////////////////////////
void DisasmWindow::DiscardDeviceResources()
{
	SafeRelease(&pRenderTarget);
	SafeRelease(&pBlackBrush);
}


////////////////////////////////////////////////////////////////////////////////////////////////////
//
// 1ߕ̏Zbg
//
////////////////////////////////////////////////////////////////////////////////////////////////////
void DisasmWindow::SetOperation(OperationSet* oSet)
{
	int bytesCount;
	int opCode;
	vector<uint8_t> byteList;
	wchar_t bytesString[30] = {};
	
	opCode = sys.pAddress->ReadByte(curAddress);
	bytesCount = bytes[opCode];
	oSet->startAddress = curAddress;

	for (int i = 0; i < bytesCount; ++i)
	{
		wchar_t buf[20] = {};
		wsprintf(buf, L"%02X ", (int)sys.pAddress->ReadByte(curAddress));
		wcscat(bytesString, buf);
		byteList.push_back(sys.pAddress->ReadByte(curAddress));
		++curAddress;
	}
	wcscat(bytesString, L"          ");
	wcsncpy(oSet->bytesString, bytesString, 10);
	oSet->bytesString[10] = 0;
	for (unsigned int i = 0; i < byteList.size(); ++i) {
		oSet->byteArray[i] = byteList[i];
	}
	wcscpy(oSet->nimonic, nimonic[opCode]);
	wchar_t str[20] = {};
	MakeOperand(str, byteList, addressing[opCode], bytes[opCode]);
	wcscpy(oSet->operand, str);
}


////////////////////////////////////////////////////////////////////////////////////////////////////
//
// Iyh쐬
//
////////////////////////////////////////////////////////////////////////////////////////////////////
void DisasmWindow::MakeOperand(wchar_t* s, vector<uint8_t>& ops, int amode, int bytes)
{
	wchar_t w[10] = {};

	switch (amode)
	{
	case (int)AddressingMode::IMMEDIATE:
		if (bytes == 1)
			wcscpy(s, L"");
		else if (bytes == 2) {
			wchar_t w[10] = {};
			wsprintf(w, L"#%02X", ops[1]);
			wcscpy(s, w);
		}
		else {
			wsprintf(w, L"#%02X%02X", ops[1], ops[2]);
			wcscpy(s, w);
		}
		break;
	case (int)AddressingMode::DIRECT:
	{
		wsprintf(w, L"%02X", ops[1]);
		wcscpy(s, w);
		break;
	}
	case (int)AddressingMode::INDEX:
	{
		wsprintf(w, L"%02X,X", ops[1]);
		wcscpy(s, w);
		break;
	}
	case (int)AddressingMode::EXTEND:
	{
		wsprintf(w, L"%02X%02X", ops[1], ops[2]);
		wcscpy(s, w);
		break;
	}
	case (int)AddressingMode::RELATIVEAD:
	{

		int i = (int8_t)(ops[1]) + curAddress;
		wsprintf(w, L"%02X", (uint16_t)i);
		wcscpy(s, w);
		break;
	}
	}
}


////////////////////////////////////////////////////////////////////////////////////////////////////
//
// EBhEvV[W
//
////////////////////////////////////////////////////////////////////////////////////////////////////
LRESULT CALLBACK WndProcDisasmWindow(HWND hWnd, UINT message, WPARAM wParam, LPARAM lParam)
{
	LRESULT result = 0;
	bool wasHandled = false;
	RECT rect;

	switch (message) {
	case WM_SIZE:
	{
		UINT w = LOWORD(lParam);
		UINT h = HIWORD(lParam);
		g_pDisasmWindow->OnResize(w, h);

		GetClientRect(hWnd, &rect);
		dispno = rect.bottom / lineHeight;
		break;
	}
	case WM_PAINT:
	case WM_DISPLAYCHANGE:
	{
		PAINTSTRUCT ps;
		BeginPaint(hWnd, &ps);
		g_pDisasmWindow->OnRender();
		EndPaint(hWnd, &ps);
		break;
	}
	case WM_DESTROY:
		g_bDisasmWindow = false;
		CheckMenuItem(GetMenu(g_hMainWnd), IDM_TOOL_DISASSEMBLEWINDOW, MF_UNCHECKED);
		delete g_pDisasmWindow;
		break;

	default:
		return DefWindowProc(hWnd, message, wParam, lParam);
	}
	return 0;
}


